#!/bin/sh

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


GEN=./gen

# Generate docs for SynTest test cases

$GEN 0 > incfile0.txt
$GEN 100 > incfile1.txt
$GEN 50000 > incfile2.txt

# Generate docs for SDKTest test cases

# docs of size 0, 10k, 20k, 30k, ... 90k
for i in 0 1 2 3 4 5 6 7 8 9
do
  $GEN ${i}0000 > file${i}.txt
done

# docs of size 1.1k, 1.2k, ...,2.0k, ..., 9.9k
for i in 1 2 3 4 5 6 7 8 9
do
    for j in 0 1 2 3 4 5 6 7 8 9
    do
        $GEN ${i}${j}00 > file${i}${j}.txt
    done
done
